import requests
from .objects.accounts import Accounts
from .objects.customfields import CustomFields
from .objects.customredirects import CustomRedirects
from .objects.dynamiccontent import DynamicContent
from .objects.emailclicks import EmailClicks
from .objects.emailtemplates import EmailTemplates
from .objects.forms import Forms
from .objects.lifecyclehistories import LifecycleHistories
from .objects.lifecyclestages import LifecycleStages
from .objects.lists import Lists
from .objects.listmemberships import ListMemberships
from .objects.emails import Emails
from .objects.prospects import Prospects
from .objects.opportunities import Opportunities
from .objects.prospectaccounts import ProspectAccounts
from .objects.tags import Tags
from .objects.tagobjects import TagObjects
from .objects.users import Users
from .objects.visits import Visits
from .objects.visitors import Visitors
from .objects.visitoractivities import VisitorActivities
from .objects.campaigns import Campaigns

from .errors import PardotAPIError

# Issue #1 (http://code.google.com/p/pybing/issues/detail?id=1)
# Python 2.6 has json built in, 2.5 needs simplejson
try:
    import json
except ImportError:
    import simplejson as json

BASE_URI = 'https://pi.pardot.com'


class PardotAPI(object):
    def __init__(self, email, password, user_key, version=4):
        self.email = email
        self.password = password
        self.user_key = user_key
        self.api_key = None
        self.version = version
        self.accounts = Accounts(self)
        self.campaigns = Campaigns(self)
        self.customfields = CustomFields(self)
        self.customredirects = CustomRedirects(self)
        self.dynamiccontent = DynamicContent(self)
        self.emailclicks = EmailClicks(self)
        self.emails = Emails(self)
        self.emailtemplates = EmailTemplates(self)
        self.forms = Forms(self)
        self.lifecyclehistories = LifecycleHistories(self)
        self.lifecyclestages = LifecycleStages(self)
        self.listmemberships = ListMemberships(self)
        self.lists = Lists(self)
        self.opportunities = Opportunities(self)
        self.prospects = Prospects(self)
        self.prospectaccounts = ProspectAccounts(self)
        self.tags = Tags(self)
        self.tagobjects = TagObjects(self)
        self.users = Users(self)
        self.visits = Visits(self)
        self.visitors = Visitors(self)
        self.visitoractivities = VisitorActivities(self)

    def post(self, object_name, path=None, params=None, retries=0):
        """
        Makes a POST request to the API. Checks for invalid requests that raise PardotAPIErrors. If the API key is
        invalid, one re-authentication request is made, in case the key has simply expired. If no errors are raised,
        returns either the JSON response, or if no JSON was returned, returns the HTTP response status code.
        """
        if params is None:
            params = {}
        params.update({'user_key': self.user_key, 'api_key': self.api_key, 'format': 'json'})
        try:
            self._check_auth(object_name=object_name)
            request = requests.post(self._full_path(object_name, self.version, path), params=params)
            response = self._check_response(request)
            return response
        except PardotAPIError as err:
            if err.message == 'Invalid API key or user key':
                response = self._handle_expired_api_key(err, retries, 'post', object_name, path, params)
                return response
            else:
                raise err

    def get(self, object_name, path=None, params=None, retries=0):
        """
        Makes a GET request to the API. Checks for invalid requests that raise PardotAPIErrors. If the API key is
        invalid, one re-authentication request is made, in case the key has simply expired. If no errors are raised,
        returns either the JSON response, or if no JSON was returned, returns the HTTP response status code.
        """
        if params is None:
            params = {}
        params.update({'user_key': self.user_key, 'api_key': self.api_key, 'format': 'json'})
        try:
            self._check_auth(object_name=object_name)
            request = requests.get(self._full_path(object_name, self.version, path), params=params)
            response = self._check_response(request)
            return response
        except PardotAPIError as err:
            if err.message == 'Invalid API key or user key':
                response = self._handle_expired_api_key(err, retries, 'get', object_name, path, params)
                return response
            else:
                raise err

    def _handle_expired_api_key(self, err, retries, method, object_name, path, params):
        """
        Tries to refresh an expired API key and re-issue the HTTP request. If the refresh has already been attempted,
        an error is raised.
        """
        if retries != 0:
            raise err
        self.api_key = None
        if self.authenticate():
            response = getattr(self, method)(object_name=object_name, path=path, params=params, retries=1)
            return response
        else:
            raise err

    @staticmethod
    def _full_path(object_name, version, path=None):
        """Builds the full path for the API request"""
        full = '{0}/api/{1}/version/{2}'.format(BASE_URI, object_name, version)
        if path:
            return full + '{0}'.format(path)
        return full

    @staticmethod
    def _check_response(response):
        """
        Checks the HTTP response to see if it contains JSON. If it does, checks the JSON for error codes and messages.
        Raises PardotAPIError if an error was found. If no error was found, returns the JSON. If JSON was not found,
        returns the response status code.
        """
        if response.headers.get('content-type') == 'application/json':
            json = response.json()
            error = json.get('err')
            if error:
                raise PardotAPIError(json_response=json)
            return json
        else:
            return response.status_code

    def _check_auth(self, object_name):
        if object_name == 'login':
            return
        if self.api_key is None:
            self.authenticate()

    def authenticate(self):
        """
         Authenticates the user and sets the API key if successful. Returns True if authentication is successful,
         False if authentication fails.
        """
        try:
            auth = self.post('login', params={'email': self.email, 'password': self.password})
            self.api_key = auth.get('api_key')
            if self.api_key is not None:
                return True
            return False
        except PardotAPIError:
            return False



class ProspectAccounts(object):
    """
    A class to query and use Pardot prospect accounts.
    Prospect account field reference: http://developer.pardot.com/kb/object-field-references/#prospect-account
    """

    def __init__(self, client):
        self.client = client

    def query(self, **kwargs):
        """
        Returns the prospect accounts matching the specified criteria parameters.
        Supported search criteria: http://developer.pardot.com/kb/api-version-4/prospect-accounts/#supported-search-criteria
        """
        response = self._get(path='/do/query', params=kwargs)

        # Ensure result['prospectAccount'] is a list, no matter what.
        result = response.get('result')
        if result['total_results'] == 0:
            result['prospectAccount'] = []
        elif result['total_results'] == 1:
            result['prospectAccount'] = [result['prospectAccount']]

        return result

    def create(self, **kwargs):
        """Creates a new prospect account."""
        response = self._post(path='/do/create', params=kwargs)
        return response

    def describe(self, **kwargs):
        """
        Returns the field metadata for prospect accounts, explaining what fields are available, their types, whether
        they are required, and their options (for dropdowns, radio buttons, etc).
        """
        response = self._get(path='/do/describe', params=kwargs)
        return response

    def read(self, id=None, **kwargs):
        """
        Returns the data for the prospect account specified by <id>. <id> is the Pardot ID of the target prospect
        account.
        """
        response = self._post(path='/do/read/id/{id}'.format(id=id), params=kwargs)
        return response

    def update(self, id=None, **kwargs):
        """
        Updates the data for the prospect account specified by <id>. <id> is the Pardot ID of the target prospect
        account.
        """
        response = self._post(path='/do/update/id/{id}'.format(id=id), params=kwargs)
        return response
    
    def assign(self, id=None, user_id=None, **kwargs):
        """
        Assigns the prospect account to a user specified by <user_id>. <id> is the Pardot ID of the target prospect account.
        """
        response = self._post(path='/do/assign/id/{id}'.format(id=id), params=kwargs)
        return response

    def _get(self, object_name='prospectAccount', path=None, params=None):
        """GET requests for the Prospect Account object."""
        if params is None:
            params = {}
        response = self.client.get(object_name=object_name, path=path, params=params)
        return response

    def _post(self, object_name='prospectAccount', path=None, params=None):
        """POST requests for the Prospect Account object."""
        if params is None:
            params = {}
        response = self.client.post(object_name=object_name, path=path, params=params)
        return response



class PardotAPIError(Exception):
    """
    Basic exception class for errors encountered in API post and get requests. Takes the json response and parses out
    the error code and message.
    """

    def __init__(self, json_response):
        self.response = json_response
        self.err_code = json_response.get('@attributes').get('err_code')
        self.message = str(json_response.get('err'))
        if self.err_code is None:
            self.err_code = 0
            self.message = 'Unknown API error occurred'

    def __str__(self):
        return 'Error #{err_code}: {message}'.format(err_code=self.err_code, message=self.message)


class PardotAPIArgumentError(Exception):
    pass



class Users(object):
    """
    A class to query and use Pardot users.
    User field reference: http://developer.pardot.com/kb/object-field-references
    """

    def __init__(self, client):
        self.client = client

    def query(self, **kwargs):
        """
        Returns the users matching the specified criteria parameters.
        Supported search criteria: http://developer.pardot.com/kb/api-version-4/users/#supported-search-criteria
        """
        response = self._get(path='/do/query', params=kwargs)

        # Ensure result['users'] is a list, no matter what.
        result = response.get('result')
        if result['total_results'] == 0:
            result['user'] = []
        elif result['total_results'] == 1:
            result['user'] = [result['user']]

        return result

    def read_by_id(self, id=None, **kwargs):
        """
        Returns the data for the user specified by <id>. <id> is the Pardot ID of the target user."""
        response = self._post(path='/do/read/id/{id}'.format(id=id), params=kwargs)
        return response

    def read_by_email(self, email=None, **kwargs):
        """
        Returns the data for the user specified by <email>. <email> is the email address of the target user."""
        response = self._post(path='/do/read/email/{email}'.format(email=email), params=kwargs)
        return response

    def _get(self, object_name='user', path=None, params=None):
        """GET requests for the User object."""
        if params is None:
            params = {}
        response = self.client.get(object_name=object_name, path=path, params=params)
        return response

    def _post(self, object_name='user', path=None, params=None):
        """POST requests for the User object."""
        if params is None:
            params = {}
        response = self.client.post(object_name=object_name, path=path, params=params)
        return response



class Visits(object):
    """
    A class to query and use Pardot visits.
    Visit field reference: http://developer.pardot.com/kb/object-field-references/#visit
    """

    def __init__(self, client):
        self.client = client

    def query_by_ids(self, ids=None, **kwargs):
        """Returns the visits matching the given <ids>. The <ids> should be comma separated integers (no spaces)."""
        kwargs['ids'] = ids.replace(' ', '')
        response = self._get(path='/do/query', params=kwargs)

        # Ensure result['visit'] is a list, no matter what.
        result = response.get('result')
        if result['total_results'] == 0:
            result['visit'] = []
        elif result['total_results'] == 1:
            result['visit'] = [result['visit']]

        return result

    def query_by_visitor_ids(self, visitor_ids=None, **kwargs):
        """
        Returns the visits matching the given <visitor ids>. The <visitor ids> should be comma separated integers
        (no spaces).
        """
        kwargs['visitor_ids'] = visitor_ids.replace(' ', '')
        response = self._get(path='/do/query', params=kwargs)

        # Ensure result['visit'] is a list, no matter what.
        result = response.get('result')
        if result['total_results'] == 0:
            result['visit'] = []
        elif result['total_results'] == 1:
            result['visit'] = [result['visit']]

        return result

    def query_by_prospect_ids(self, prospect_ids=None, **kwargs):
        """
        Returns the visits matching the given <prospect ids>. The <prospect ids> should be comma separated integers
        (no spaces).
        """
        kwargs['prospect_ids'] = prospect_ids.replace(' ', '')
        response = self._get(path='/do/query', params=kwargs)

        # Ensure result['visit'] is a list, no matter what.
        result = response.get('result')
        if result['total_results'] == 0:
            result['visit'] = []
        elif result['total_results'] == 1:
            result['visit'] = [result['visit']]

        return result

    def read(self, id=None, **kwargs):
        """
        Returns the data for the visit specified by <id>. <id> is the Pardot ID of the target visit."""
        response = self._post(path='/do/read/id/{id}'.format(id=id), params=kwargs)
        return response

    def _get(self, object_name='visit', path=None, params=None):
        """GET requests for the Visit object."""
        if params is None:
            params = {}
        response = self.client.get(object_name=object_name, path=path, params=params)
        return response

    def _post(self, object_name='visit', path=None, params=None):
        """POST requests for the Visit object."""
        if params is None:
            params = {}
        response = self.client.post(object_name=object_name, path=path, params=params)
        return response



class Emails(object):
    """
    A class to query and send Pardot emails.
    Email field reference: http://developer.pardot.com/kb/object-field-references/#email
    """

    def __init__(self, client):
        self.client = client

    def send_to_email(self, prospect_email=None, **kwargs):
        """
        Sends an email to the prospect identified by <prospect_email>.
        Required parameters: (email_template_id OR (text_content, name, subject, & ((from_email & from_name) OR from_user_id)))
        """
        response = self._post(
            path='/do/send/prospect_email/{prospect_email}'.format(prospect_email=prospect_email),
            params=kwargs)
        return response

    def send_to_id(self, prospect_id=None, **kwargs):
        """
        Sends an email to the prospect identified by <prospect_id>.
        Required parameters: (email_template_id OR (text_content, name, subject, & ((from_email & from_name) OR from_user_id)))
        """
        response = self._post(
            path='/do/send/prospect_id/{prospect_id}'.format(prospect_id=prospect_id), params=kwargs)
        return response

    def send_to_lists(self, **kwargs):
        """
        Sends an email to the lists identified by list_ids[].
        Required parameters: (email_template_id OR (text_content, name, subject, & ((from_email & from_name) OR from_user_id)))
        """
        kwargs['list_ids'] = kwargs.get('list_ids', None)
        response = self._post(
            path='/do/send/', params=kwargs)
        return response

    def read(self, email_id=None):
        """Returns the data for the email specified by <email_id>. <email_id> is the Pardot ID of the target email."""
        response = self._post(path='/do/read/id/{email_id}'.format(email_id=email_id))
        return response
    
    def stats(self, list_email_id=None):
        """Returns the statistical data for the list email specified by <list_email_id>. <list_email_id> is the Pardot ID of the target email."""
        response = self._post(path='/do/read/id/{list_email_id}'.format(list_email_id=list_email_id))
        return response

    def _get(self, object_name='email', path=None, params=None):
        """GET requests for the Email object."""
        if params is None:
            params = {}
        response = self.client.get(object_name=object_name, path=path, params=params)
        return response

    def _post(self, object_name='email', path=None, params=None):
        """POST requests for the Email object."""
        if params is None:
            params = {}
        response = self.client.post(object_name=object_name, path=path, params=params)
        return response



class Opportunities(object):
    """
    A class to query and use Pardot opportunities.
    Opportunity field reference: http://developer.pardot.com/kb/object-field-references#opportunity
    """

    def __init__(self, client):
        self.client = client

    def query(self, **kwargs):
        """
        Returns the opportunities matching the specified criteria parameters.
        Supported search criteria: http://developer.pardot.com/kb/api-version-4/opportunities/#supported-search-criteria
        """
        response = self._get(path='/do/query', params=kwargs)

        # Ensure result['opportunity'] is a list, no matter what.
        result = response.get('result')
        if result['total_results'] == 0:
            result['opportunity'] = []
        elif result['total_results'] == 1:
            result['opportunity'] = [result['opportunity']]

        return result

    def create_by_email(self, prospect_email=None, name=None, value=None, probability=None, **kwargs):
        """
        Creates a new opportunity using the specified data. <prospect_email> must correspond to an existing prospect.
        """
        kwargs.update({'name': name, 'value': value, 'probability': probability})
        response = self._post(
            path='/do/create/prospect_email/{prospect_email}'.format(prospect_email=prospect_email),
            params=kwargs)
        return response

    def create_by_id(self, prospect_id=None, name=None, value=None, probability=None, **kwargs):
        """
        Creates a new opportunity using the specified data. <prospect_id> must correspond to an existing prospect.
        """
        kwargs.update({'name': name, 'value': value, 'probability': probability})
        response = self._post(
            path='/do/create/prospect_id/{prospect_id}'.format(prospect_id=prospect_id),
            params=kwargs)
        return response

    def read(self, id=None):
        """
        Returns the data for the opportunity specified by <id>, including campaign assignment and associated visitor
        activities. <id> is the Pardot ID for the target opportunity.
        """
        response = self._post(path='/do/read/id/{id}'.format(id=id))
        return response

    def update(self, id=None):
        """
        Updates the provided data for the opportunity specified by <id>. <id> is the Pardot ID for the target
        opportunity. Fields that are not updated by the request remain unchanged. Returns an updated version of the
        opportunity.
        """
        response = self._post(path='/do/update/id/{id}'.format(id=id))
        return response

    def delete(self, id=None):
        """
        Deletes the opportunity specified by <id>. <id> is the Pardot ID for the target opportunity. Returns no response
        on success.
        """
        response = self._post(path='/do/delete/id/{id}'.format(id=id))
        return response

    def undelete(self, id=None):
        """
        Un-deletes the opportunity specified by <id>. <id> is the Pardot ID for the target opportunity. Returns no
        response on success.
        """
        response = self._post(path='/do/undelete/id/{id}'.format(id=id))
        return response

    def _get(self, object_name='opportunity', path=None, params=None):
        """GET requests for the Opportunity object."""
        if params is None:
            params = {}
        response = self.client.get(object_name=object_name, path=path, params=params)
        return response

    def _post(self, object_name='opportunity', path=None, params=None):
        """POST requests for the Opportunity object."""
        if params is None:
            params = {}
        response = self.client.post(object_name=object_name, path=path, params=params)
        return response



from setuptools import setup


setup(
    name="PyPardot4",
    version="1.1.8",
    author="Matt Needham",
    author_email="matthew.m.needham@gmail.com",
	license="MIT",
    description="API wrapper for APIv4 of Pardot marketing automation software.",
    keywords="pardot",
    url="https://github.com/mneedham91/PyPardot4",
    packages=['pypardot', 'pypardot.objects'],
    install_requires=['requests'],
)
